@@ -10,7 +10,7 @@ from .common import *
 from .contentAPI import ContentAPI
 from .entityAPI import EntityAPI
 from .uploadAPI import UploadAPI, simple_asset_package, complex_asset_package, cvs_to_xsd, cvs_to_xml, \
-                    cvs_to_cmis_xslt, csv_to_search_xml, generic_asset_package, upload_config
+                    cvs_to_cmis_xslt, csv_to_search_xml, generic_asset_package, upload_config, multi_asset_package
 from .workflowAPI import WorkflowAPI, WorkflowContext, WorkflowInstance
 from .retentionAPI import RetentionAPI, RetentionAssignment, RetentionPolicy
 from .parAPI import PreservationActionRegistry
@@ -18,6 +18,6 @@ from .parAPI import PreservationActionRegistry
 __author__ = "James Carr (drjamescarr@gmail.com)"
 
 # Version of the Preservica API package
-__version__ = "1.0.2"
+__version__ = "1.1.1"
 
 __license__ = "Apache License Version 2.0"
