@@ -1,6 +1,8 @@
 """
 Base class for authenticated API calls used by Entity, Content and Upload
 
+Manages the authentication token lifetime and namespace versions.
+
 author:     James Carr
 licence:    Apache License 2.0
 
@@ -14,8 +16,6 @@ import threading
 import time
 import xml.etree.ElementTree
 from enum import Enum
-from typing import Union
-
 import requests
 import logging
 import unicodedata
@@ -30,6 +30,7 @@ CHUNK_SIZE = 1024 * 2
 NS_XIP_ROOT = "http://preservica.com/XIP/"
 NS_ENTITY_ROOT = "http://preservica.com/EntityAPI/"
 NS_RM_ROOT = "http://preservica.com/RetentionManagement/"
+NS_SEC_ROOT = "http://preservica.com/SecurityAPI"
 
 NS_WORKFLOW = "http://workflow.preservica.com"
 
@@ -68,7 +69,8 @@ class FileHash:
 
 
 def _make_stored_zipfile(base_name, base_dir, owner, group, verbose=0, dry_run=0, zlogger=None):
-    """Create a zip file from all the files under 'base_dir'.
+    """
+    Create a non compressed zip file from all the files under 'base_dir'.
 
     The output zip file will be named 'base_name' + ".zip".  Returns the
     name of the output zip file.
@@ -117,7 +119,7 @@ class PagedSet:
     The results object contains the list of objects of interest
     """
 
-    def __init__(self, results, has_more, total, next_page):
+    def __init__(self, results, has_more: bool, total: int, next_page: str):
         self.results = results
         self.has_more = bool(has_more)
         self.total = int(total)
@@ -173,6 +175,50 @@ class UploadProgressCallback:
             sys.stdout.flush()
 
 
+class RelationshipDirection(Enum):
+    FROM = "From"
+    TO = "To"
+
+
+class EntityType(Enum):
+    ASSET = "IO"
+    FOLDER = "SO"
+    CONTENT_OBJECT = "CO"
+
+
+class Relationship:
+    DCMI_hasFormat = "http://purl.org/dc/terms/hasFormat"
+    DCMI_isFormatOf = "http://purl.org/dc/terms/isFormatOf"
+    DCMI_hasPart = "http://purl.org/dc/terms/hasPart"
+    DCMI_isPartOf = "http://purl.org/dc/terms/isPartOf"
+    DCMI_hasVersion = "http://purl.org/dc/terms/hasVersion"
+    DCMI_isVersionOf = "http://purl.org/dc/terms/isVersionOf"
+    DCMI_isReferencedBy = "http://purl.org/dc/terms/isReferencedBy"
+    DCMI_references = "http://purl.org/dc/terms/references"
+    DCMI_isReplacedBy = "http://purl.org/dc/terms/isReplacedBy"
+    DCMI_replaces = "http://purl.org/dc/terms/replaces"
+    DCMI_isRequiredBy = "http://purl.org/dc/terms/isRequiredBy"
+    DCMI_requires = "http://purl.org/dc/terms/requires"
+    DCMI_conformsTo = "http://purl.org/dc/terms/conformsTo"
+
+    def __init__(self, relationship_id: str, relationship_type: str, direction: RelationshipDirection, other_ref: str,
+                 title: str, entity_type: EntityType, this_ref: str, api_id: str):
+        self.api_id = api_id
+        self.this_ref = this_ref
+        self.entity_type = entity_type
+        self.title = title
+        self.other_ref = other_ref
+        self.direction = direction
+        self.relationship_type = relationship_type
+        self.relationship_id = relationship_id
+
+    def __str__(self):
+        if self.direction == RelationshipDirection.FROM:
+            return f"{self.this_ref} {self.relationship_type} {self.other_ref}"
+        else:
+            return f"{self.other_ref} {self.relationship_type} {self.this_ref}"
+
+
 class IntegrityCheck:
     """
     Class to hold information about completed integrity checks
@@ -359,12 +405,6 @@ class Thumbnail(Enum):
     LARGE = "large"
 
 
-class EntityType(Enum):
-    ASSET = "IO"
-    FOLDER = "SO"
-    CONTENT_OBJECT = "CO"
-
-
 def sanitize(filename):
     """
     Return a fairly safe version of the filename.
@@ -418,6 +458,42 @@ class AuthenticatedAPI:
     Base class for authenticated calls which need an access token
     """
 
+    def security_tags_base(self, with_permissions: bool = False):
+        """
+             Return available security tags
+
+             :return: dict of security tags
+             :rtype:  dict
+         """
+
+        if (self.major_version < 7) and (self.minor_version < 4) and (self.patch_version < 1):
+            raise RuntimeError("security_tags API call is only available with a Preservica v6.3.1 system or higher")
+
+        headers = {HEADER_TOKEN: self.token, 'Content-Type': 'application/xml;charset=UTF-8'}
+
+        request = self.session.get(f'https://{self.server}/api/security/tags', headers=headers)
+        if request.status_code == requests.codes.ok:
+            xml_response = str(request.content.decode('utf-8'))
+            logger.debug(xml_response)
+            entity_response = xml.etree.ElementTree.fromstring(xml_response)
+            security_tags = dict()
+            tags = entity_response.findall(f'.//{{{self.sec_ns}}}Tag')
+            for tag in tags:
+                permissions = list()
+                for p in tag.findall(f'.//{{{self.sec_ns}}}Permission'):
+                    permissions.append(p.text)
+                if with_permissions:
+                    security_tags[tag.attrib['name']] = permissions
+                else:
+                    security_tags[tag.attrib['name']] = tag.attrib['name']
+            return security_tags
+        if request.status_code == requests.codes.unauthorized:
+            self.token = self.__token__()
+            return self.security_tags_base()
+        else:
+            logger.error(f'security_tags failed {request.status_code}')
+            raise RuntimeError(request.status_code, "security_tags failed")
+
     def entity_from_string(self, xml_data: str):
         entity_response = xml.etree.ElementTree.fromstring(xml_data)
         reference = entity_response.find(f'.//{{{self.xip_ns}}}Ref')
@@ -451,12 +527,14 @@ class AuthenticatedAPI:
                 self.xip_ns = f"{NS_XIP_ROOT}v{self.major_version}.{self.minor_version}"
                 self.entity_ns = f"{NS_ENTITY_ROOT}v{self.major_version}.{self.minor_version}"
                 self.rm_ns = f"{NS_RM_ROOT}v{self.major_version}.{2}"
+                self.sec_ns = f"{NS_SEC_ROOT}/v{self.major_version}.{self.minor_version}"
 
     def __version_number__(self):
         """
         Determine the version number of the server
         """
         headers = {HEADER_TOKEN: self.token}
+        self.version_hash = "RHVOMzJHNzdk"
         request = self.session.get(f'https://{self.server}/api/entity/versiondetails/version', headers=headers)
         if request.status_code == requests.codes.ok:
             xml_ = str(request.content.decode('utf-8'))
@@ -504,14 +582,13 @@ class AuthenticatedAPI:
     def __token__(self):
         logger.debug("Token Expired Requesting New Token")
         if self.shared_secret is False:
-            if self.tenant == "%":
+            if self.tenant is None:
                 data = {'username': self.username, 'password': self.password, 'includeUserDetails': 'true'}
             else:
                 data = {'username': self.username, 'password': self.password, 'tenant': self.tenant}
-            logger.error(data)
             response = self.session.post(f'https://{self.server}/api/accesstoken/login', data=data)
             if response.status_code == requests.codes.ok:
-                if self.tenant == "%":
+                if self.tenant is None:
                     self.tenant = response.json()['tenant']
                 return response.json()['token']
             else:
@@ -535,10 +612,10 @@ class AuthenticatedAPI:
                 logger.error(msg)
                 raise RuntimeError(response.status_code, msg)
 
-    def __init__(self, username: str = None, password: str = None, tenant: str = "%", server: str = None,
+    def __init__(self, username: str = None, password: str = None, tenant: str = None, server: str = None,
                  use_shared_secret: bool = False):
         config = configparser.ConfigParser()
-        config.read('credentials.properties')
+        config.read('credentials.properties', encoding='utf-8')
         self.session = requests.Session()
         self.shared_secret = bool(use_shared_secret)
 
@@ -570,7 +647,7 @@ class AuthenticatedAPI:
         else:
             self.password = password
 
-        if not tenant or tenant == "%":
+        if not tenant:
             tenant = os.environ.get('PRESERVICA_TENANT')
             if tenant is None:
                 try:
@@ -579,10 +656,8 @@ class AuthenticatedAPI:
                     pass
         if not tenant:
             msg = "No valid tenant found in method arguments, environment variables or credentials.properties file"
-            logger.error(msg)
-            raise RuntimeError(msg)
-        else:
-            self.tenant = tenant
+            logger.debug(msg)
+        self.tenant = tenant
 
         if not server:
             server = os.environ.get('PRESERVICA_SERVER')
