@@ -1,9 +1,11 @@
+import base64
 import csv
 import json
 import shutil
 import tempfile
 import uuid
 import xml
+from pathlib import Path
 from time import sleep
 
 import boto3
@@ -19,6 +21,9 @@ from botocore.exceptions import ClientError
 
 import s3transfer.upload
 import s3transfer.tasks
+import cryptography
+from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
+from requests.auth import HTTPBasicAuth
 from s3transfer import S3UploadFailedError
 
 from pyPreservica.common import *
@@ -91,8 +96,10 @@ def prettify(elem):
     return re_parsed.toprettyxml(indent="  ")
 
 
-def __create_io__(file_name=None, parent_folder=None, **kwargs):
-    xip = Element('XIP')
+def __create_io__(xip=None, file_name=None, parent_folder=None, **kwargs):
+    if xip is None:
+        xip = Element('XIP')
+    assert xip is not None
     xip.set('xmlns', 'http://preservica.com/XIP/v6.0')
     io = SubElement(xip, 'InformationObject')
     ref = SubElement(io, 'Ref')
@@ -657,6 +664,137 @@ def generic_asset_package(preservation_files_dict=None, access_files_dict=None,
         return top_level_folder + ".zip"
 
 
+def multi_asset_package(asset_file_list=None, export_folder=None, parent_folder=None, compress=True, **kwargs):
+    """
+    Create a package containing multiple assets, all the assets are ingested into the same parent folder provided
+    by the parent_folder argument.
+
+    :param asset_file_list: List of files. One asset per file
+    :param export_folder:   Location where the package is written to
+    :param parent_folder:   The folder the assets will be ingested into
+    :param compress:        Bool, compress the package
+    :param kwargs:
+    :return:
+    """
+
+    # some basic validation
+    if export_folder is None:
+        export_folder = tempfile.gettempdir()
+    if not os.path.isdir(export_folder):
+        logger.error("Export Folder Does Not Exist")
+        raise RuntimeError(export_folder, "Export Folder Does Not Exist")
+    if parent_folder is None:
+        logger.error("You must specify a parent folder for the package asset")
+        raise RuntimeError("You must specify a parent folder for the package asset")
+
+    security_tag = kwargs.get('SecurityTag', "open")
+    content_type = kwargs.get('CustomType', "")
+
+    if not compress:
+        shutil.register_archive_format("szip", _make_stored_zipfile, None, "UnCompressed ZIP file")
+
+    if 'Preservation_files_fixity_callback' in kwargs:
+        fixity_callback = kwargs.get('Preservation_files_fixity_callback')
+    else:
+        fixity_callback = Sha1FixityCallBack()
+
+    package_id = str(uuid.uuid4())
+    export_folder = export_folder
+    top_level_folder = os.path.join(export_folder, package_id)
+    os.mkdir(top_level_folder)
+    inner_folder = os.path.join(top_level_folder, package_id)
+    os.mkdir(inner_folder)
+    os.mkdir(os.path.join(inner_folder, "content"))
+
+    asset_map = dict()
+    xip = Element('XIP')
+    for file in asset_file_list:
+        default_asset_title = os.path.splitext(os.path.basename(file))[0]
+        xip, io_ref = __create_io__(xip, file_name=default_asset_title, parent_folder=parent_folder, **kwargs)
+        asset_map[file] = io_ref
+        representation = SubElement(xip, 'Representation')
+        io_link = SubElement(representation, 'InformationObject')
+        io_link.text = io_ref
+        access_name = SubElement(representation, 'Name')
+        access_name.text = "Preservation"
+        access_type = SubElement(representation, 'Type')
+        access_type.text = "Preservation"
+        content_objects = SubElement(representation, 'ContentObjects')
+        content_object = SubElement(content_objects, 'ContentObject')
+        content_object_ref = str(uuid.uuid4())
+        content_object.text = content_object_ref
+
+        default_content_objects_title = os.path.splitext(os.path.basename(file))[0]
+        content_object = SubElement(xip, 'ContentObject')
+        ref_element = SubElement(content_object, "Ref")
+        ref_element.text = content_object_ref
+        title = SubElement(content_object, "Title")
+        title.text = default_content_objects_title
+        description = SubElement(content_object, "Description")
+        description.text = default_content_objects_title
+        security_tag_element = SubElement(content_object, "SecurityTag")
+        security_tag_element.text = security_tag
+        custom_type = SubElement(content_object, "CustomType")
+        custom_type.text = content_type
+        parent = SubElement(content_object, "Parent")
+        parent.text = io_ref
+
+        generation = SubElement(xip, 'Generation', {"original": "true", "active": "true"})
+        content_object = SubElement(generation, "ContentObject")
+        content_object.text = content_object_ref
+        label = SubElement(generation, "Label")
+        label.text = os.path.splitext(os.path.basename(file))[0]
+        effective_date = SubElement(generation, "EffectiveDate")
+        effective_date.text = datetime.now().isoformat()
+        bitstreams = SubElement(generation, "Bitstreams")
+        bitstream = SubElement(bitstreams, "Bitstream")
+        bitstream.text = os.path.basename(file)
+        SubElement(generation, "Formats")
+        SubElement(generation, "Properties")
+
+        bitstream = SubElement(xip, 'Bitstream')
+        filename_element = SubElement(bitstream, "Filename")
+        filename_element.text = os.path.basename(file)
+        filesize = SubElement(bitstream, "FileSize")
+        file_stats = os.stat(file)
+        filesize.text = str(file_stats.st_size)
+        physical_location = SubElement(bitstream, "PhysicalLocation")
+        fixities = SubElement(bitstream, "Fixities")
+        fixity_result = fixity_callback(filename_element.text, file)
+        if type(fixity_result) == tuple:
+            fixity = SubElement(fixities, "Fixity")
+            fixity_algorithm_ref = SubElement(fixity, "FixityAlgorithmRef")
+            fixity_value = SubElement(fixity, "FixityValue")
+            fixity_algorithm_ref.text = fixity_result[0]
+            fixity_value.text = fixity_result[1]
+        elif type(fixity_result) == dict:
+            for key, val in fixity_result.items():
+                fixity = SubElement(fixities, "Fixity")
+                fixity_algorithm_ref = SubElement(fixity, "FixityAlgorithmRef")
+                fixity_value = SubElement(fixity, "FixityValue")
+                fixity_algorithm_ref.text = key
+                fixity_value.text = val
+        else:
+            logger.error("Could Not Find Fixity Value")
+            raise RuntimeError("Could Not Find Fixity Value")
+
+        src_file = file
+        dst_file = os.path.join(os.path.join(inner_folder, "content"), os.path.basename(file))
+        shutil.copyfile(src_file, dst_file)
+
+    if xip is not None:
+        metadata_path = os.path.join(inner_folder, "metadata.xml")
+        metadata = open(metadata_path, "wt", encoding='utf-8')
+        metadata.write(prettify(xip))
+        metadata.close()
+        if compress:
+            shutil.make_archive(top_level_folder, 'zip', top_level_folder)
+        else:
+            shutil.make_archive(top_level_folder, 'szip', top_level_folder)
+        shutil.rmtree(top_level_folder)
+        return top_level_folder + ".zip"
+
+
 def complex_asset_package(preservation_files_list=None, access_files_list=None, export_folder=None,
                           parent_folder=None, compress=True, **kwargs):
     """
@@ -942,6 +1080,10 @@ def upload_config():
     return transfer_config
 
 
+def _unpad(s):
+    return s[:-ord(s[len(s) - 1:])]
+
+
 class UploadAPI(AuthenticatedAPI):
 
     def ingest_twitter_feed(self, twitter_user=None, num_tweets: int = 25, twitter_consumer_key=None,
@@ -1211,9 +1353,98 @@ class UploadAPI(AuthenticatedAPI):
 
             self.upload_zip_package(path_to_zip_package=package, folder=parent_folder, callback=callback)
 
+    def __convert_(self, key, cypher_text):
+        base64_decoded = base64.b64decode(cypher_text)
+        key = base64.b64decode(self.version_hash.encode("utf-8") + key.encode("UTF-8")).decode("utf-8").encode("utf-8")
+        aes = cryptography.hazmat.primitives.ciphers.algorithms.AES(key)
+        cipher = Cipher(algorithm=aes, mode=modes.ECB())
+        decryptor = cipher.decryptor()
+        output_bytes = decryptor.update(base64_decoded) + decryptor.finalize()
+        return _unpad(output_bytes.decode("utf-8")).strip()
+
+    def upload_buckets(self):
+        """
+        Get a list of available upload buckets
+
+        :return: dict of bucket names and regions
+        """
+        request = self.session.get(f"https://{self.server}/api/admin/locations/upload",
+                                   auth=HTTPBasicAuth(self.username, self.password))
+
+        buckets = dict()
+        xml_tag = "N2YxcGVsUA=="
+        if request.status_code == requests.codes.ok:
+            xml_response = str(request.content.decode('utf-8'))
+            entity_response = xml.etree.ElementTree.fromstring(xml_response)
+            data_sources = entity_response.findall('.//dataSource')
+            for data_source in data_sources:
+                sip_locations = data_source.findall('.//sipLocation')
+                for sip_location in sip_locations:
+                    buckets[self.__convert_(xml_tag, sip_location.text)] = self.__convert_(xml_tag, sip_location.attrib[
+                        'region'])
+        return buckets
+
+    def upload_zip_package_to_S3(self, path_to_zip_package, bucket_name, folder=None, callback=None,
+                                 delete_after_upload=False):
+
+        """
+         Uploads a zip file package to an S3 bucket connected to a Preservica Cloud System
+
+         :param str path_to_zip_package: Path to the package
+         :param str bucket_name: Bucket connected to an ingest workflow
+         :param Folder folder: The folder to ingest the package into
+         :param str callback: Optional callback to allow the callee to monitor the upload progress
+         :param bool delete_after_upload: Delete the local copy of the package after the upload has completed
+
+        """
+
+        request = requests.get(f"https://{self.server}/api/admin/locations/upload?refresh={bucket_name}",
+                               auth=HTTPBasicAuth(self.username, self.password))
+
+        if request.status_code is not requests.codes.ok:
+            raise SystemError(request.content)
+        if request.status_code == requests.codes.ok:
+            xml_response = str(request.content.decode('utf-8'))
+            entity_response = xml.etree.ElementTree.fromstring(xml_response)
+            a = entity_response.find('.//a')
+            b = entity_response.find('.//b')
+            c = entity_response.find('.//c')
+            xml_tag = "N2YxcGVsUA=="
+            access_key = self.__convert_(xml_tag, a.text)
+            secret_key = self.__convert_(xml_tag, b.text)
+            session_token = self.__convert_(xml_tag, c.text)
+
+            session = boto3.Session(aws_access_key_id=access_key, aws_secret_access_key=secret_key,
+                                    aws_session_token=session_token)
+            s3 = session.resource(service_name="s3")
+
+            upload_key = str(uuid.uuid4())
+            s3_object = s3.Object(bucket_name, upload_key)
+            metadata = dict()
+            metadata['key'] = upload_key
+            metadata['name'] = upload_key + ".zip"
+            metadata['bucket'] = bucket_name
+            metadata['status'] = 'ready'
+
+            if hasattr(folder, "reference"):
+                metadata['collectionreference'] = folder.reference
+            elif isinstance(folder, str):
+                metadata['collectionreference'] = folder
+
+            metadata['size'] = str(Path(path_to_zip_package).stat().st_size)
+            metadata['createdby'] = self.username
+
+            metadata_map = {'Metadata': metadata}
+
+            s3_object.upload_file(path_to_zip_package, Callback=callback, ExtraArgs=metadata_map,
+                                  Config=transfer_config)
+
+            if delete_after_upload:
+                os.remove(path_to_zip_package)
+
     def upload_zip_package(self, path_to_zip_package, folder=None, callback=None, delete_after_upload=False):
         """
-        Uploads a zip file package and starts an ingest workflow
+        Uploads a zip file package directly to Preservica and starts an ingest workflow
 
         :param str path_to_zip_package: Path to the package
         :param Folder folder: The folder to ingest the package into
@@ -1231,6 +1462,7 @@ class UploadAPI(AuthenticatedAPI):
         bucket = f'{self.tenant.lower()}.package.upload'
         endpoint = f'https://{self.server}/api/s3/buckets'
         self.token = self.__token__()
+
         s3_client = boto3.client('s3', endpoint_url=endpoint, aws_access_key_id=self.token,
                                  aws_secret_access_key="NOT_USED",
                                  config=Config(s3={'addressing_style': 'path'}))
@@ -1261,4 +1493,5 @@ class UploadAPI(AuthenticatedAPI):
                 return response['ResponseMetadata']['HTTPHeaders']['preservica-progress-token']
 
             except ClientError as e:
+                logger.error(e)
                 raise e
